"""This module contains methods to create a grid"""
# IMPORTING NAMESPACES:
# STANDARD LIBRARY
import numpy as np
import warnings


class Grid(object):
    """
    This class defines a grid for one-dimensional
    quantum chemisty calculations.

    Attributes
    ----------
    xmin : float
       Lower coordinate range.
    xmax : float
       Upper coordinate range.
    gpd : int
       GridPointDensity of coordinate grid.
    """

    def __init__(self):
        """ Grid constructor. """
        self.xmin = -20
        self.xmax = 20
        self.gpd = 150
        self.xcoord = []
        self.spacing = None

    def set_xrange(self, xmin, xmax):
        """ Sets the Atributes xmin, xmax."""
        self.xmin = xmin
        self.xmax = xmax
        if len(self.xcoord) != 0:
            warnings.warn("The parameters xmin and xmas were changed"
                          + " after creating a grid", stacklevel=2)

    def set_grid_point_density(self, gpd):
        """ Sets the grid point density. """
        self.gpd = gpd
        if len(self.xcoord) != 0:
            warnings.warn("The parameters gpd was changed "
                          + " after creating a grid", stacklevel=2)

    def show_attributes(self, show=False):
        """
        Shows Grid's current attributes.

        Parameters
        ----------
        show : Bool
            If True xcoords will be printed.
        """
        print('\n\n---------------')
        print('1D-GRID DETAILS')
        print('---------------')
        print('Lower coordinate range: xmin = ', self.xmin)
        print('Upper coordinate range: xmax = ', self.xmax)
        print('Grid point density:     gpd  = ', self.gpd)
        print('Evenly spaced Grid:            ', self.spacing)
        xcoord = self.xcoord
        if show:
            print('\nCoordinates: xcoord = \n', xcoord)
        lencoord = len(xcoord)
        if len(xcoord) != 0:
            if self.xmin != xcoord[0] or self.xmax != xcoord[lencoord-1]:
                warnings.warn("Warning: The set x-range parameters "
                              + " do not match the given coordinates!",
                              stacklevel=2)
            current_gpd = int(len(xcoord)/abs(xcoord[lencoord-1]-xcoord[0]))
            if self.gpd != current_gpd:
                warnings.warn("Warning: The set gpd do not match "
                              + " with the given coordinates!", stacklevel=2)

    def create_linear_grid(self):
        """
        Creates a linear grid space with evenly spaced numbers.

        Returns
        -------
        x : float (real numbers)
           real linear coordinate space.
        """
        if len(self.xcoord) != 0:
            warnings.warn("A grid already exists.", stacklevel=2)
        # define number of gridpoints with gpd
        deltaX = abs(self.xmax-self.xmin)
        ngp = int(deltaX*self.gpd)
        x = np.linspace(self.xmin, self.xmax, num=ngp)
        self.xcoord = x
        self.spacing = True
        return x

    def set_default(self):
        """ Sets all attributes to default."""
        self.xmin = -20
        self.xmax = 20
        self.gpd = 150
        self.xcoord = []
        self.spacing = None

    def delete(self):
        """ delets all attributes to None or empty lists."""
        self.xmin = None
        self.xmax = None
        self.gpd = None
        self.xcoord = []
        self.spacing = None


def customized_linear_grid(xmin=-20, xmax=20, gpd=150, show=False):
    """
    creates a customized linear (evenly spaced)
    one dimensinal grid (x-coordinates).

    Parameters
    ----------
    xmin : float
       Lower coordinate range.
    xmax : float
       Upper coordinate range.
    gpd : int
       GridPointDensity of coordinate grid.
    Returns
    -------
    x : float (real numbers)
       real linear coordinate space.
    """
    grid = Grid()
    grid.set_xrange(xmin, xmax)
    grid.set_grid_point_density(gpd)
    x = grid.create_linear_grid()
    grid.show_attributes(show=show)
    return x



""" This is the creator method for the QLine package."""

from QLine.Constants import *
from QLine.Grid import *
from QLine.basis_functions import *
from QLine.math_methods import *
from QLine.potentials import *
from QLine.hamiltonian import *
from QLine.schroedinger_eq import *
from QLine.plotting import *

import os
import QLine
path = os.path.dirname(QLine.__file__)


print("=========================================")
print("        Python showcase package")
print("                QLine  ")
print("             Version: 1.0            2023")
print("-----------------------------------------")
print("Author: Michael Welzel")
print("location: ", path)
print("License: GNU GENERAL PUBLIC LICENSE")
print("-----------------------------------------")
print("Everything is calculated in atomic units")
print('=========================================')



"""
This module creates the schroedinger equation and
solves this  eigenvalue equation via the basis set method.
"""

# IMPORTING NAMESPACES:
# PACKAGE MODULES
import QLine.math_methods as mm
import QLine.hamiltonian as ha
import QLine.basis_functions as bf
import QLine.Constants as const
# STANDARD LIBRARY
import numpy as np
import heapq


def basis_set_method(x, pot, N=10, m=1, kb=1, mb=1,
                     pre_check=True, periodic=False):
    """
    Solves the Schroedinger equation using the variational principle
    (basis set method):

    Hc = lamb*c

    Parameters
    ----------
    x : numpy.array
       x-coordinates (corresponding to the potential function).
    pot : numpy.array
       potential function range.
    N : int
       basis size and hamiltonian dimension(NxN).
    m : float
       particle mass
    kb : float; Real>0
       wave number of basis functions (harmonic oscillators).
    mb : float
       particle mass of basis functions (harmonic oscillators).

    Returns
    -------
    lamb : array
       energy eigenvalues.
    coeff : ndarray
       coefficient matrix.
    basis : ndarray[:,i]
       set of used basis functions.
    """
    # Crating basis set:
    bs = bf.Basis_Set(x, N)
    basis, basis2dev = bs.create_harmonic_oscillators_vary_n(m=mb, k=kb)
    # Creating hamiltonian:
    matrix = ha.matrix(x, basis, pre_check=pre_check, periodic=periodic)
    H = matrix.hamiltonian(basis2dev, pot, m=m)
    # Solving eigenvalue equation
    lamb, coeff = np.linalg.eigh(H)
    return lamb, coeff, basis


def wave_function_from_coeff(coeff: np.ndarray,
                             basis: np.ndarray) -> np.ndarray:
    """
    composes the coefficients and the basis function to
    the corresponding wavefunction Hpsi=Epsi with

    psi = sum_ij c[ij] * basis[:j]

    Parameters
    ----------
    coeff : ndarray
       coefficient matrix of solving Schroedinger equation(basis_set_method).
    basis : ndarray(dim(xcoord) x N)
        Set of basis functions.

    Returns
    -------
    wafu : ndarray[:,i]
       wave functions of state i.
    """
    N = len(basis[0, :])
    wafu = np.zeros((len(basis[:, 0]), int(N)), dtype=complex)
    for i in range(N):
        for j in range(N):
            wafu[:, i] = wafu[:, i] + coeff[j, i]*basis[:, j]
    return wafu


class Discretization_method(object):
    """
    Solves the one-dimensional Schroedinger equation
    via the discretization method
    (didactic verision - not optimized for perfomance).

    Parameters
    ----------
    x : numpy.array
       One-dimensional coordinates (real) corresponding to the given potential.
    pot : numpy.array
       function value of a given potential function.

    Returns
    -------
    H : numpy.ndarray
       Hamiltonian matrix.
    """

    def __init__(self, x):
        """class constructor."""
        self.x = x
        self.dim = len(x)

    def kinetic(self, m=1):
        """creates the descretized kinetical energy part
           of 1D-Schroedinger equation.

        Parameters
        ----------
        m : float>0
           particle mass.

        Returns
        -------
        T : numpy.ndarray
            kinetical part of hamiltonian.
        """
        self.m = m
        d = abs(self.x[1]-self.x[0])  # grid distance
        term = (-const.hbar**2)/(2*m*d**2)
        T = mm._sec_derivative_matrix(self.dim)
        T = term*T
        self.T = T
        return T

    def potential(self, pot):
        """creates the descretized potential energy part
           of 1D-Schroedinger equation.

        Parameters
        ----------
        pot : numpy.array
           function value of a given potential function.

        Returns
        -------
        V : numpy.ndarray
            potential part of hamiltonian.
        """
        ones = np.identity(self.dim)
        V = pot*ones
        self.pot = pot
        self.V = V
        return V

    def hamiltonian(self, pot, m=1, solve=False):
        """creates the descretized hamiltonian
           of 1D-Schroedinger equation.

        Parameters
        ----------
        pot : numpy.array
           function value of a given potential function.
        m : float>0
           particle mass.
        solve : bool
           If True: solve Hamiltonian eigenvalue problem

        Returns
        -------
        H : numpy.ndarray
            potential part of Hamiltonian.
        energies : numpy.ndarray
           energie eigenvalues.
        wavefuncs : numpy.ndarray
           corresponding wave functions. wavefuncs[:,index]
        """
        T = self.kinetic(m=m)
        V = self.potential(pot)
        H = T + V
        self.H = H
        if solve:
            eigval, eigvec = np.linalg.eig(self.H)
            # sorting eigenvalues and vectoers
            index_ew = heapq.nsmallest(self.dim, enumerate(eigval),
                                       key=lambda x: x[1])
            energies = np.zeros(self.dim, dtype=complex)
            wavefuncs = np.zeros((self.dim, self.dim), dtype=complex)
            for i in range(self.dim):
                energies[i] = eigval[index_ew[i][0]]  # Sorted eigenvalues
                # Sorted - eigenvectors (coloumn vectors)
                wavefuncs[:, i] = eigvec[:, index_ew[i][0]]
            self.energies = energies.real
            # normalize wave functions
            for i in range(self.dim):
                norm = mm.norm(self.x, wavefuncs[:, i])
                wavefuncs[:, i] = wavefuncs[:, i]/norm
            self.wavefuncs = wavefuncs
            return energies.real, wavefuncs
        else:
            return H



"""This module contains methods to create basis functions and a basis set."""

# IMPORTING NAMESPACES:
# PACKAGE MODULES
import QLine.Constants as const
# STANDARD LIBRARY
import math as ma
import numpy as np


class Basis_Functions(object):
    """ Class of basis functions.

    Attributes
    ----------
    xcoord : np.ndarray real
       x-coordinates in one-dimension.
    psi : np.ndarray  complex
       Function values of the basis function to
       the corresponding given x-coodinates.
    d2_psi : np.ndarray complex
       Second derivative of the corresponding psi function.
    """

    def __init__(self, xcoord):
        """
        Basis_Functions constructor
        - For Details see Basis_Function docstring.
        """
        self.xcoord = xcoord
        self.psi = []
        self.d2_psi = []

    def create_harmonic_oscillator(self, v, m=1, k=1, show=False):
        """
        Creates harmonic oscillator functions.

        Parameters
        ----------
        v : int
           Quantum number.
        m : float; Real>0
           Particle mass.
        k : float; Real>0
           wave number.

        Returns
        -------
        Psi : float; array(Real)
        Harmonic oscillator function values (physically scaled).
        """
        if show:
            print('\n\n--- Harmonic Oscillator Attributes ---')
            print('       Quantum number, v = ', v)
            print('       Particle mass,  m = ', m)
            print('       Wavenumber,     k = ', k)
            print('--------------------------------------')
        # Normalizations and scalings:
        alph = np.sqrt(k*m/const.hbar)  # Norm scale term
        norm = 1/(np.sqrt(2**float(v)*ma.factorial(v)))  # norm of function
        chi = alph**0.5*self.xcoord  # scaling parameter for hermite polynom
        pre_term = (alph/ma.pi)**0.25  # Expanded Pre-Term
        # Needed Functions:

        def _gauss(z):
            """Gaussian."""
            gauss = np.exp(-z**2/2)  # gauss function
            return gauss

        def _hermite_polynom(z, n):
            """
            Creates hermite polynoms (physical version).

            Parameters
            ----------
            z : float ; Real
               (Physically) Scaled coordinates.
            n : int
               Hermite polynom order.

            Returns
            -------
            H : array(1 x dim(x-coordinates))
            Functionsvalues of a hermite polynom in order n.
            """
            # Hermite polynomial - recursive formula
            H = [np.ones(len(z)), 2.*z]
            for i in range(1, n+1):
                Hn = 2.*z*H[i]-2.*i*H[i-1]
                H.append(Hn)
            return H[n]

        hermit = _hermite_polynom(chi, v)  # Hermite Polynomfunction
        gauss = _gauss(chi)
        psi = norm*pre_term*hermit*gauss  # Harmonic oscillator
        self.psi = psi
        return psi

    def create_sec_derivative_of_harmonic_oscillator(self, v, m=1, k=1):
        """
        Creates the analytical second derivative of the harmonic oscillator.

        For attributes, parameters and returns see at
        create_harmonic_oscillator().
        """
        dim = len(self.xcoord)
        alph = np.sqrt(k*m/const.hbar)  # Pre-Term
        chi = alph**0.5*self.xcoord  # scaling parameter for hermite polynom
        ones = np.ones(dim)
        term = chi**2 - ones - 2*ones*v
        ps = self.create_harmonic_oscillator(v, m=m, k=k)
        d2_psi = term*ps*alph
        self.d2_psi = d2_psi
        self.psi = ps
        return d2_psi


class Basis_Set(object):
    """
    Creates a basis set.

    Attributes
    ----------
    xcoord : array
       x-coordinates.
    N : int
       Basis set size.
    basis_set : ndarray(dim(xcoord) x N)
        Set of basis functions.
    deriv_basis_set : ndarray(dim(xcoord) x N)
        Second derivatives of the corresponding basis functions.
    analytic_deriv : Bool
        Second derivative is analytical(True)/numerical(False)
    """
    def __init__(self, xcoord, N):
        """Basis_Set constructor - For Details see Basis_Set docstring."""
        self.xcoord = xcoord
        self.N = int(N)
        self.basis_set = []
        self.deriv2_basis_set = []
        self.analytic_deriv = None

    def create_harmonic_oscillators_vary_n(self, m=1, k=1, analytical=True):
        """
        Creates a base set consisting of harmonic oscillators
        (and their seond derivatives) ordered by quantum number.

        Parameters
        ----------
        m : float; Real>0
           Particle mass.
        k : float; Real>0
           wave number.
        analytical : Bool
           Second derivative is analytical (True) or numerical (False)

        Returns
        -------
        basis_set[:,v] : float; ndarray(dim(xcoord x N)), Real
           Set of harmonic oscillator functions (physically scaled).
        deriv2_basis_set[:,v] : float; ndarray(dim(xcoord x N)), Real
           Set of second derivatives of corresponding
           harmonic oscillator functions.
        """
        self.analytic_deriv = analytical
        dim_coord = len(self.xcoord)
        # Create basis functions
        self.basis_set = np.zeros((dim_coord, self.N))
        BF = Basis_Functions(self.xcoord)
        for v in range(self.N):
            self.basis_set[:, v] = BF.create_harmonic_oscillator(v, m=m, k=k)
        # Create second derivatives of basis functions
        self.deriv2_basis_set = np.zeros((dim_coord, self.N))
        if analytical:  # Analytical
            for v in range(self.N):
                self.deriv2_basis_set[:, v] = (
                   BF.create_sec_derivative_of_harmonic_oscillator(v, m=m, k=k)
                )
        return self.basis_set, self.deriv2_basis_set



"""This module contains several potential functions."""
# IMPORTING NAMESPACES:
# STANDARD LIBRARY
import numpy as np
import math as ma
# PACKAGE MODULES
import QLine.Constants as const


def harmonic_potential(x, k=1):
    """Creates a harmonic potential.

    Parameters
    ----------
    k : float
       harmonic force constant.
    x : ndarray
       coordinates.

    Returns
    -------
    v : ndarray
    Harmonic potential function (range).
    """
    v = 0.5*k*x**2
    return v


def harmonic_potential_mass_freq(x, m=1, w=1):
    """Creates a harmonic potential.

    Parameters
    ----------
    w : float>0
       frequency
    x : ndarray
       coordinates.

    Returns
    -------
    v : ndarray
    Harmonic potential function (range).
    """
    v = 0.5*m*w**2*x**2
    return v


def force_constant_harmonic_osc(m, w):
    """
    calculates the harmonic oscillator force constant
    based on the mass m and frequency w.
    k = mw^2

    Parameters
    ----------
    m : float>0
       particle mass
    w : float>0
       frequency

    Returns
    -------
    k : float>0
       force constant
    """
    k = m*w**2
    return k


def frequency_harmonic_osc(m, k):
    """
    calculates the harmonic oscillator frequency
    based on the mass m and frequency w.
    w = sqrt(k/m)

    Parameters
    ----------
    m : float>0
       particle mass
    k : float>0
       force constant

    Returns
    -------
    w : float>0
       frequency
    """
    w = np.sqrt(k/m)
    return w


def energy_harmonic_osc(m, w, N):
    """
    calculates the anlytical energy of
    a harmonic oscillator.
    En  = hbar*w(0.5+n)

    Parameters
    ----------
    m : float>0
       particle mass
    w : float>0
       frequency
    N : int
       Number of energies to be displayed starting from the lowest one.

    Returns
    -------
    E : list(float)
       List of energies of the harmonic oscillator.
    """
    E = []
    for i in range(N):
        Ei = const.hbar*w*(i+0.5)
        E.append(Ei)
    return E


def morse_potential(x, De, ke=1, x0=0):
    """ Creates a morse potential.

    Parameters
    ----------
    x : ndarray
       coorinates.
    De : float
       Potential well depth.
    ke : float
       Morse force constant.
    x0 : float
       point of equilibrium.

    Returns
    -------
    v : ndarray
    Morse potential function (range).
    """
    a = ma.sqrt(ke/(2.*De))
    t1 = np.exp(-2.*a*(x-x0))
    t2 = np.exp(-a*(x-x0))
    v = De*(t1-2.*t2)
    return v



"""This module includes various useful and mathematical methods."""
# IMPORTING NAMESPACES:
# STANDARD LIBRARY
import numpy as np
import scipy.linalg as sp  # science python
import warnings


def even_space_check(x: np.ndarray) -> bool:
    """ Checks if an array is evenly spaced.

    Parameters
    ----------
    x : array
       ordinary array.

    Returns
    -------
    check : bool
       evenly spaced(True), not evenly spaced(False)
    """
    dim = len(x)
    dm = []
    for i in range(dim-1):
        dm.append(x[i+1]-x[i])
    dmtest = np.full((1, len(x)-1), x[1]-x[0])
    check = np.allclose(dm, dmtest)
    return check


def _sec_derivative_matrix(dim, formula='fast'):
    """
    creates the second derivative matrix:
    (-2   1   0        ...   0 )
    ( 1  -2   1  0     ...   0 )
    ( 0   1  -2  1  0  ...   0 )

    ( ........................ )

    ( 0 ......... 0  1  -2   1 )
    ( 0 ............ 0   1  -2 )

    without the term 1/Delta^2.

    Parameters
    ----------
    dim : int
       dimension of the matrix

    Returns
    -------
    kinmat : numpy.ndarray
        kinetic matrix [dimension x dimension]
    """
    if formula == 'education':
        kinmat = np.zeros((dim, dim), dtype=complex)
        for i in range(dim):
            for j in range(dim):
                if i == j:
                    kinmat[i, j] = -2
                elif i == j + 1 or i == j - 1:
                    kinmat[i, j] = 1
    else:
        two = -2*np.ones(dim)
        ones = np.ones(dim-1)
        kinmat = (np.diagflat(two, 0)
                  + np.diagflat(ones, 1)
                  + np.diagflat(ones, -1))
    return kinmat


def second_derivative_num(x: np.ndarray, func: np.ndarray) -> np.ndarray:
    """
    Creates the second derivative using the matrix-vector formalism
    of an function with compact support (function approaches
    zero at domain edges) and a evenly spaced grid.

    finite difference method: 3-point-formalism.

    Parameters
    ----------
    x : array
       coordinates - Domain space of a function.
    func : array
       functionvalues - image space of a function.

    Returns
    -------
    d2func : array
       numerical second derivative of the given function.
    """
    check = even_space_check(x)
    if check is False:
        raise ValueError("Coordinates are not evenly spaced.")
    check = function_range_threshhold_check(func)
    if check is False:
        str1 = "The input function is non-zero on the input boundaries. "
        warn_size = str1 + "Edge of the derivation possibly unreasonable"
        warnings.warn(warn_size, stacklevel=2)
    dim = len(x)
    d = abs(x[1]-x[0])
    d2 = _sec_derivative_matrix(dim)
    d2func = np.matmul(d2, func)/d**2.  # matrix vector multiplication
    return d2func


def braket(x_coord: np.ndarray, f1: np.ndarray, f2: np.ndarray) -> np.ndarray:
    # Scalarproduct of functions/ Bra-Ket
    """
    calculates the scalar product <f,g>, where
    f,g are functions.
    and <.,.> is the sesquilinear inner product.
    The function domain and range have to be evenly spaced.

    Parameters
    ----------
    x_coord : array
        coordinates (function range of f1,f2).
    f1,f2 : array[:]
        function range/image.

    Returns
    -------
    scapr : float
    scalar product of the functions f1,f2 with their domain x_coord.
    """
    check = even_space_check(x_coord)
    if check is False:
        raise ValueError("The parameter x_coord have to be evenly spaced.")
    d = np.abs(x_coord[1]-x_coord[0])
    scapr = np.vdot(f1, f2)*d  # d = integration measure
    return scapr


def norm(x_coord: np.ndarray, f: np.ndarray) -> np.ndarray:
    # Scalarproduct  norm of functions/ Bra-Ket
    """
    calculates the scalar product norm
           ||f|| = sqrt(<f,f>),
    where f is a function.
    and <.,.> is the sesquilinear inner product.
    The function domain and range have to be evenly spaced.

    Parameters
    ----------
    x_coord : array
        coordinates (function range of f).
    f : array[:]
        function range/image.

    Returns
    -------
    scapr : float
    scalar product of the functions f1,f2 with their domain x_coord.
    """
    check = even_space_check(x_coord)
    if check is False:
        raise ValueError("The parameter x_coord have to be evenly spaced.")
    d = np.abs(x_coord[1]-x_coord[0])
    scapr = np.sqrt(np.vdot(f, f)*d)  # d = integration measure
    return scapr


def _overlapmatrix(x_coord, basis):
    """
    calculates the overlap matrix S=<fi,fj>:
    fi = basis[:,i],
    fj = basis[:,j],
    where x_coord -> fi,fj.

    Parameters
    ----------
    x_coord : numpy.array, real
       one-dimensional coordinates.
    basis : numpry.ndarray
        matrix of "i" basis functions [:,i]

    Returns
    -------
    S : numpy.ndarray
       overlap matrix.
    """
    N = len(basis[0, :])
    S = np.zeros((N, N), dtype=complex)
    for i in range(N):
        for j in range(N):
            f1 = basis[:, i]
            f2 = basis[:, j]
            S[i, j] = braket(x_coord, f1, f2)
    return S


def orthogonalization_loewdin(x_coord, basis, show=False):
    """
    calculates a loewdin orthogonalization for a given basis set
    to build a reasonable ONS.

    Parameters
    ----------
    x_coord : numpy.array, real
       one-dimensional coordinates.
    basis : numpry.ndarray
        matrix of "i" basis functions [:,i]
    show : Bool
        if True : overlap and transformation matrix will be printed.

    Returns
    -------
    Q : numpy.ndarray
       ortogonlizated basis.
    T : numpy.ndarray
       Transformation Matrix.
    """
    print("\n\n"+"---"*15+"\n Loewdin orthogonalization algorithm is used.\n")

    # Create overlap matrix
    S = _overlapmatrix(x_coord, basis)

    # Create Transformation matrix T
    def _loewdin_transformation(S):
        """
        Transformation: Q = XT = X S^(-1/2)
        X = given basis set
        T = transformation matrix
        S = overlap matrix
        Q = orthonormalized basis set
        """
        S_sqr = sp.sqrtm(S)
#        S_inv_sqr = np.linalg.inv(S_sqr)
        S_inv_sqr = sp.inv(S_sqr)
        T = S_inv_sqr
        return T

    T = _loewdin_transformation(S)

    # Orthonormalize given Basis
    Q = np.matmul(basis, T)

    # show intermediate steps
    if show:
        print("\nOverlap matrix: S= \n", np.around(S, decimals=3))
        print("\nTransformation matrix:T=S^(-1/2)= \n",
              np.around(T, decimals=3))
    return Q, T


def function_range_threshhold_check(func):
    """
    checks if the boundary values of a given function are approximately zero.

    Parameters
    ----------
    func : numpy.array
       Arbitrary function values.

    Returns
    -------
    check : bool
       True if the function is zero at the edges.
    """
    thresh = 0
    start = func[0]
    end = func[len(func)-1]
    check_start = np.isclose(start, thresh)
    check_end = np.isclose(end, thresh)
    if check_start and check_end:
        check = True
    else:
        check = False
    return check



"""
This module creates the schroedinger equation and
solves this  eigenvalue equation via the basis set method.
"""

# STANDARD LIBRARY
import numpy as np
import matplotlib.pylab as plt


def plot_wfunctions_within_potential(xcoord: np.ndarray, pot: np.ndarray,
                                     lamb: np.ndarray, wavefunc: np.ndarray,
                                     xmin=-10, xmax=10, ymin=-3, ymax=12,
                                     savefig=False
                                     ) -> None:
    """
    Plots the wave functions within the corresponding potential.

    Parameters
    ----------
    xcoord : array
       x-coordinates.
    pot : arrray
       potential function range/image
    wavefunc : ndarray[:,i]
       wavefunction ranges from Schroedinger equation.
    lamb : array
       energy eigenvalues.
    xmin : float
       lower threshold for plot in respect to the coordinate x.
    xmax : float
       uper threshold for plot in respect to the coordinate x.
    ymin : float
       lower threshold for plot in respect to the function values(potential).
    ymax : float
       upper threshold for plot in respect to the function values(potential).
    savefig : bool
       if true = Figure will be saved as png
    """
    plt.figure()
    plt.axhline(y=0, color="black", linestyle="--", lw=2)
    plt.plot(xcoord, pot, lw=4, label=r"V(x)")
    for i, val in enumerate(lamb):
        wfplot = wavefunc[:, i].real + val
        plt.axhline(y=val, color="black", linestyle="--", lw=1)
        plt.plot(xcoord, wfplot, color="blue")
    plt.plot(xcoord, wfplot, color="blue", label=r"$\psi_n(x)$")
    plt.xlim(xmin, xmax)
    plt.ylim(ymin, ymax)
    plt.legend()
    plt.xlabel("x/a.u.")
    plt.ylabel("E/a.u.")
    if savefig:
        plt.savefig("HO_pot_wf.png")
    plt.show()


def plot_energy_levels(lamb: np.ndarray, decimals=1, savefig=False) -> None:
    """
    plots the energy levels.

    Parameters
    ----------
    lamb : array
        energy eigenvalues.
    decimals : int
       Number of decimal places in the plot.
    savefig : bool
       if true = Figure will be saved as png
    """
    plt.figure()
    plt.axhline(y=0, color="purple", linestyle="--", lw=2)
    for _i, val in enumerate(lamb):
        plt.axhline(y=val, xmin=0.2, xmax=0.3,
                    color="black", linestyle="-", lw=2)
        plt.text(0.16, val, np.around(val, decimals=decimals))
    plt.xticks([])
    plt.yticks(fontsize=20)
    plt.xlim(0, 0.5)
    plt.xlabel("E/a.u.")
    plt.title("Energy level")
    if savefig:
        plt.savefig("HO_Elevel.png")
    plt.show()



import setuptools
from setuptools import setup, find_packages

with open("README.md", "r") as f:
    ldescription = f.read()

setuptools.setup(
    include_package_data = True,
    name         = 'QLine',
    version      = '1.2',
    description  = 'For the calculation of one-dimensional Schroedinger equations',
    author       = 'Michael Welzel',
    license      = 'GPLv3',
    package_dir  = {'': 'src/'},
    python_requires = '>=3.11.4',
    install_requires = ['numpy>=1.24.2','matplotlib>=3.7.0', 'scipy>=1.10.1'],
    classifiers  = ["Programming Language :: Python :: 3",
                    "Operating System :: OS Independent"],
    long_description=ldescription,
    long_description_content_type='text/markdown',  
    keywords=["python", "quantum", "mechanics", "one", "dimensional", "chemistry",
              "harmonic", "oscillator", "basis set method", "variatonal principle"
             ],
    )



"""This module creates the hamiltonian and correspondings matrices."""

# IMPORTING NAMESPACES:
# PACKAGE MODULES
import QLine.math_methods as mm
# STANDARD LIBRARY
import numpy as np
import warnings


class matrix(object):
    """
    Creates a matrix object with Matrix:=<X|O|X">,
    where <.,.> is a scalar product, O an Operator,
    X the basis function and X" the second derivative
    of the basis function.
    The basis set have to be orthonormalized set.

    Parameters
    ----------
    coord : np.ndarray
       Coordinates 1D.
    basis_set : np.ndarray[:,i]
        basis functions, numpy.array[:,i]; where i is the function set
        index number.
    N : int
       matrix dimensions and basis set size.
    pre_check : bool
       If True - checks the input data for correctness
       in respect to the basis set.
    loew_T : numpy.ndarray or None
       loewdin orthonormalization transformation matrix
       (see src/math_methods.py).
    periodic : bool
       if True: basis functions have not to reach zero
       at coordinate limits(periodic basis functions).
    """
    def __init__(self, x_coord, basis_set, pre_check=True, periodic=False):
        """matrix constructor - For Details see matrix docstring."""
        self.x_coord = x_coord
        self.basis_set = basis_set
        self.N = len(basis_set[0, :])
        self.loew_T = None

        # checking the input data for correctness:
        self.pre_check = pre_check
        if self.pre_check:

            # Dimensional check for basis set input
            if len(basis_set[0, :]) >= len(basis_set[:, 0]):
                warn_size = ("Dimensions of your basis functions"
                             + " could be arranged incorrectly.")
                warnings.warn(warn_size, stacklevel=2)

            # Verifying that basis set approch zero at long distances
            if periodic is False:
                for i in range(self.N):
                    check = mm.function_range_threshhold_check(basis_set[:, i])
                    if check is False:
                        raise ValueError("The basis set functions must reach"
                                         + " zero at coordinate limits.")

            # verifying:  orthonormality and linear dependency
            S = self.overlap()
            # linear dependency
            det = np.linalg.det(S)
            print('det=', det)
            if np.isclose(det, 0):
                raise ValueError("The given basis set is "
                                 + "linearly dependend.")
            # Orthonormality
            ref_ONB = np.identity(self.N)
            if np.allclose(abs(S), abs(ref_ONB)) is False:
                warn_ONB = ("The given basis is not orthonormalized and thus "
                            + "unreasonable results are possible.")
                warnings.warn(warn_ONB, stacklevel=2)
                print("If you do not want to perform orthonormalization, "
                      + "suppress the pre-check:")
                print('matrix(x_coord,basis_set,pre_check=False)')
                print('                         ---------------')
                # try loewdin orthonormalization:
                Q, T = mm.orthogonalization_loewdin(self.x_coord,
                                                    self.basis_set)
                self.basis_set = Q
                self.loew_T = T
                print('Basis set Matrix       [dim(x) times N] : B')
                print('Orthonormalized Matrix [dim(x) times N] : Q')
                print('Transformation matrix       [N times N] : trans')
                print('-----------------------------------------------')
                print('Q = B*trans')

    def overlap(self, show=False) -> np.ndarray:
        """
        Calculates the overlapp matrix S_ij = <f_i|f_j>,
        where f_i are the basis functions with index i and j.

        Parameters
        ----------
        show : bool
           If True - printing overlap/orthonormality matrix.

        Returns
        -------
        S_mat : ndarray[N,N]
        overlap matrix.
        """
        S_mat = mm._overlapmatrix(self.x_coord, self.basis_set)
        if show:
            print("\nOrthonormality/overlap matrix: S=\n",
                  np.around(S_mat, decimals=2))
        return S_mat

    def kinetic(self, sec_derivative: np.ndarray,
                m=1, show=False) -> np.ndarray:
        """
        Creates the matrix of kinetic energy Tij=(-1/(2m))<f_i,f"_j>,
        where f_i are the basis functions with index i and j.
        f" is the corresponding second derivative.

        Parameters
        ----------
        sec_derivative : ndarray[:,i]
             have to be the set of second derivatives of the given basis set.
        m : float
            particle mass.
        show : bool
           If True - printing overlap/orthonormality matrix.

        Returns
        -------
        T : ndarray[N,N]
        Matrix of kinetic energy.
        """
        N = self.N
        if np.shape(self.loew_T) == (N, N):
            sec_derivative = np.matmul(sec_derivative, self.loew_T)
            str1_1 = "2nd derivatives of the given (untransformed) "
            str1_2 = "basis functions were transformed "
            str1 = str1_1 + str1_2
            str2 = "according to loewdin-orthonormalization which which has "
            str3 = "already been performed for the given basis.\n"
            print("\nFor your information:\n"+str1+str2+str3)
            print("2nd derivative basis set Matrix       "
                  + "[dim(x) times N] : B\'\'")
            print("Orthonormalized 2nd derivative Matrix "
                  + "[dim(x) times N] : Q\'\'")
            print("Transformation matrix       [N times N] : trans")
            print("----------------------------------------------------------")
            print("Q\'\' = B\'\'*trans")
        T = np.zeros((N, N), dtype=complex)
        pre_factor = (-1.)/(2.*m)
        for i in range(N):
            for j in range(N):
                f1 = self.basis_set[:, i]
                f2 = sec_derivative[:, j]
                T[i, j] = mm.braket(self.x_coord, f1, f2)*pre_factor
        if show:
            print("\nKinetic matrix: T=\n", np.around(T, decimals=2))
        return T

    def potential(self, pot: np.ndarray, show=False) -> np.ndarray:
        """
        Creates the matrix of potential energy Vij=<f_i,V*f_j>,
        where f_i are the basis functions with index i and j.
        V is the potential function.

        Parameters
        ----------
        pot : array
           Potential function (range/image).
        show : bool
           If True - printing overlap/orthonormality matrix.

        Returns
        -------
        V : ndarray[N,N]
        Matrix of potential energy.
        """
        N = self.N
        V = np.zeros((N, N), dtype=complex)
        dim = len(self.basis_set[:, 0])
        pot_basis = np.zeros((dim, N), dtype=complex)
        for j in range(N):
            pot_basis[:, j] = self.basis_set[:, j]*pot[:]
        for i in range(N):
            for j in range(N):
                f1 = self.basis_set[:, i]
                f2 = pot_basis[:, j]
                V[i, j] = mm.braket(self.x_coord, f1, f2)
        if show:
            print("\nPotential matrix: V=\n", np.around(V, decimals=2))
        return V

    def hamiltonian(self, sec_derivative, pot, m=1, show=False):
        """
        Creates the hamiltonian Hij= Tij + Vij based on
        the matrix of kinetic energy Tij and potential energy Vij.

        Parameters
        ----------
        sec_derivative : ndarray[:,i]
             have to be the set of second derivatives of the given basis set.
        m : float
            particle mass.
        pot : array
           Potential function (range/image).
        show : bool
           If True - printing overlap/orthonormality matrix.

        Returns
        -------
        H : ndarray[N,N]
        Hamiltonian matrix.
        """
        T = self.kinetic(sec_derivative, m=m, show=show)
        V = self.potential(pot, show=show)
        H = T + V
        if show:
            print('\n\nHamiltonian: H=\n', np.around(H, decimals=2))
        return H
